/**
 * \file estack.c
 * \brief Rough EStack PoC for Nspire.
 * Copyright (C) 2003, 2004, 2010, 2011 Lionel Debroux (TICT).
 *
 * This is an implementation of Aitken's delta^2 algorithm, with non-interactive
 * input and non-persistent output. In the current state of things, we don't yet
 * have enough documentation on the Phoenix internals and CAS integration
 * (though there are definitely some similarities to TI-68k/AMS) to do much
 * better than this.
 *
 * We need to find how to access variables (maybe we need to embed native code
 * into regular documents ?), how to be accessible from a "Calculator" screen,
 * etc. And of course, we need to two-way reverse-engineer and reimplement this
 * silly 0D compression+encryption method.
 * But all of that is hard...
 * 
 * This code is designed to be used in conjunction with the documentation of
 * TIFS (TI's SDK for TI-68k/AMS) or GCC4TI.
 *
 * WARNING WARNING Absolute addresses are for OS 1.7.2741 CAS WARNING WARNING
 * WARNING WARNING  CRASH will occur on any other OS version  WARNING WARNING
 *
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#include <os.h>

// Addresses for OS 1.7.2741 CAS.

// AMS tags are 8 bits, but Phoenix tags are 16 bits.
typedef uint16_t ESQ;
typedef uint16_t Quantum;
typedef ESQ * ESI;
typedef const ESQ * CESI;

#define top_estack (*(ESQ **)0x1081EF80)

#define next_expression_index ((ESI (*)(CESI))0x1054A2A0)
#define remaining_element_count ((int (*)(CESI))0x105BE720)
#define compare_expressions ((int (*)(CESI, CESI))0x1054B4CC)
#define push_expression ((void (*)(CESI))0x1054A52C)

#define delete_between ((void (*)(ESI, ESI))0x1054A1C4)
#define deleted_between ((unsigned int (*)(ESI, ESI))0x1054A20C)
#define delete_expression ((void (*)(ESI))0x1054A508)
#define deleted_expression ((unsigned int (*)(ESI))0x1054A4E4)

#define Integer1Index (*(ESQ **)0x1081EF14)

#define push_quantum ((void (*)(uint16_t))0x1054A0F4)
#define push_quantum_as_nonnegative_int ((void (*)(ESQ))0x1051D198)
#define push_negate_quantum_as_negint ((void (*)(ESQ))0x1051D61C)

#define push_zstr ((void (*)(CESI))0x1054CA74)
#define push_string ((void (*)(CESI))0x105CD33C)

#define push0 ((void (*)(void))0x1051CF40)
#define push1 ((void (*)(void))0x1051CF18)

#define push_sum ((void (*)(CESI, CESI))0x1055790C)
#define push_difference ((void (*)(CESI, CESI))0x10557358)
#define push_product ((void (*)(CESI, CESI))0x105B8A68)
#define push_ratio ((void (*)(CESI, CESI))0x105BA9CC)

#define add_to_top ((void (*)(CESI))0x10558B0C)
#define subtract_from_top ((void (*)(CESI))0x1055A05C)
#define times_top ((void (*)(CESI))0x105BA520)
#define divide_top ((void (*)(CESI))0x105BA3DC)
#define and_onto_top ((void (*)(CESI))0x10540EE0)
#define or_onto_top ((void (*)(CESI))0x1054187C)

#define replace_top2_with_sum ((void (*)(ESI))0x1055850C)
#define replace_top2_with_difference ((void (*)(ESI))0x1055A088)
#define replace_top2_with_prod ((void (*)(ESI))0x105BA07C)
#define replace_top2_with_ratio ((void (*)(ESI))0x105BA280)
#define replace_top2_with_pow ((void (*)(ESI))0x10560188)
#define replace_top_with_reciprocal ((void (*)(ESI))0x105B7F3C)

#define push_parse_text ((void (*)(const char *))0x105540C8)
#define push_internal_simplify ((void (*)(CESI))0x10529C2C)
#define push_approx ((void (*)(CESI))0x10569F48)

#define push_sequence ((void (*)(CESI, CESI, CESI, CESI, CESI))0x105CC9BC)

// 441 elements in OS 1.7.2741 CAS
enum Tags {
    VAR_NAME_TAG=0x00, _VAR_Q_TAG=0x01, VAR_R_TAG=0x02, VAR_S_TAG=0x03, VAR_T_TAG=0x04, VAR_U_TAG=0x05, VAR_V_TAG=0x06, VAR_W_TAG=0x07,    VAR_X_TAG=0x08, VAR_Y_TAG=0x09, VAR_Z_TAG=0x0A, VAR_A_TAG=0x0B, VAR_B_TAG=0x0C, VAR_C_TAG=0x0D, VAR_D_TAG=0x0E, VAR_E_TAG=0x0F,    VAR_F_TAG=0x10, VAR_G_TAG=0x11, VAR_H_TAG=0x12, VAR_I_TAG=0x13, VAR_J_TAG=0x14, VAR_K_TAG=0x15, VAR_L_TAG=0x16, VAR_M_TAG=0x17,    VAR_N_TAG=0x18, VAR_O_TAG=0x19, VAR_P_TAG=0x1A, VAR_Q_TAG=0x1B,
    // ?
    ARB_REAL_TAG=0x1D, ARB_INT_TAG=0x1E,
    POSINT_TAG=0x1F, NEGINT_TAG=0x20, POSFRAC_TAG=0x21, NEGFRAC_TAG=0x22,
    FLOAT_TAG=0x23,
    PI_TAG=0x24,
    EXP_TAG=0x25,
    IM_TAG=0x26,
    NEGINFINITY_TAG=0x27, INFINITY_TAG=0x28, PN_INFINITY_TAG=0x29,
    UNDEF_TAG=0x2A, FALSE_TAG=0x2B, TRUE_TAG=0x2C,
    STR_TAG=0x2D,
    NOTHING_TAG=0x2E,
    ACOSH_TAG=0x2F, ASINH_TAG=0x30, ATANH_TAG=0x31, ASECH_TAG=0x32, ACSCH_TAG=0x33, ACOTH_TAG=0x34, COSH_TAG=0x35, SINH_TAG=0x36, TANH_TAG=0x37, SECH_TAG=0x38, CSCH_TAG=0x39, COTH_TAG=0x3A, ACOS_TAG=0x3B, ASIN_TAG=0x3C, ATAN_TAG=0x3D, ASEC_TAG=0x3E, ACSC_TAG=0x3F, ACOT_TAG=0x40, RACOS_TAG=0x41, RASIN_TAG=0x42, RATAN_TAG=0x43, COS_TAG=0x44, SIN_TAG=0x45, TAN_TAG=0x46, SEC_TAG=0x47, CSC_TAG=0x48, COT_TAG=0x49, ITAN_TAG=0x4A,
    ABS_TAG=0x4B, ANGLE_TAG=0x4C,
    CEILING_TAG=0x4D, FLOOR_TAG=0x4E, INT_TAG=0x4F, SIGN_TAG=0x50,
    SQRT_TAG=0x51,
    EXPF_TAG=0x52, LN_TAG=0x53, LOG_TAG=0x54,
    FPART_TAG=0x55, IPART_TAG=0x56,
    CONJ_TAG=0x57, IMAG_TAG=0x58, REAL_TAG=0x59,
    APPROX_TAG=0x5A,
    TEXPAND_TAG=0x5B, TCOLLECT_TAG=0x5C,
    GETDENOM_TAG=0x5D, GETNUM_TAG=0x5E,
    DELTA_LIST_TAG=0x5F,
    CUMSUM_TAG=0x60,
    DET_TAG=0x61, COLNORM_TAG=0x62, ROWNORM_TAG=0x63, NORM_TAG=0x64,
    MEAN_TAG=0x65, MEDIAN_TAG=0x66, PRODUCT_TAG=0x67, STDDEV_TAG=0x68, SUM_TAG=0x69, VARIANCE_TAG=0x6A,
    UNITV_TAG=0x6B,
    DIM_TAG=0x6C, MAT2LIST_TAG=0x6D, NEWLIST_TAG=0x6E,
    RREF_TAG=0x6F, REF_TAG=0x70,
    IDENTITY_TAG=0x71,
    DIAG_TAG=0x72, COLDIM_TAG=0x73, ROWDIM_TAG=0x74, TRANSPOSE_TAG=0x75,
    FACTORIAL_TAG=0x76,
    PERCENT_TAG=0x77,
    RADIANS_TAG=0x78,
    NOT_TAG=0x79,
    MINUS_TAG=0x7A,
    VEC_POLAR_TAG=0x7B, VEC_CYLIND_TAG=0x7C, VEC_SPHERE_TAG=0x7D,
    START_TAG=0x7E,
    ISTORE_TAG=0x7F, STORE_TAG=0x80,
    WITH_TAG=0x81,
    XOR_TAG=0x82, OR_TAG=0x83, AND_TAG=0x84,
    LT_TAG=0x85, LE_TAG=0x86, EQ_TAG=0x87, GE_TAG=0x88, GT_TAG=0x89, NE_TAG=0x8A,
    ADD_TAG=0x8B, ADDELT_TAG=0x8C,
    SUB_TAG=0x8D, SUBELT_TAG=0x8E,
    MUL_TAG=0x8F, MULELT_TAG=0x90,
    DIV_TAG=0x91, DIVELT_TAG=0x92,
    POW_TAG=0x93, POWELT_TAG=0x94,
    SINCOS_TAG=0x95,
    SOLVE_TAG=0x96, CSOLVE_TAG=0x97, NSOLVE_TAG=0x98, ZEROS_TAG=0x99, CZEROS_TAG=0x9A,
    FMIN_TAG=0x9B, FMAX_TAG=0x9C,
    COMPLEX_TAG=0x9D,
    POLYEVAL_TAG=0x9E, RANDPOLY_TAG=0x9F,
    CROSSP_TAG=0xA0, DOTP_TAG=0xA1,
    GCD_TAG=0xA2, LCM_TAG=0xA3, MOD_TAG=0xA4, INTDIV_TAG=0xA5, REMAIN_TAG=0xA6,
    NCR_TAG=0xA7, NPR_TAG=0xA8,
    P2RX_TAG=0xA9, P2RY_TAG=0xAA, P2PTHETA_TAG=0xAB, P2PR_TAG=0xAC,
    AUGMENT_TAG=0xAD, NEWMAT_TAG=0xAE, RANDMAT_TAG=0xAF, SIMULT_TAG=0xB0,
    ROOT_TAG=0xB1,
    EXP2LIST_TAG=0xB2,
    // ?
    MROW_TAG=0xB4, ROWADD_TAG=0xB5, ROWSWAP_TAG=0xB6,
    ARCLEN_TAG=0xB7, NINT_TAG=0xB8,
    PI_PRODUCT_TAG=0xB9, SIGMA_SUM_TAG=0xBA,
    MROWADD_TAG=0xBB,
    IMPDIF_TAG=0xBC,
    STDEVPOP_TAG=0xBD,
    EXACT_TAG=0xBE, // ?
    LOGB_TAG=0xBF,
    COMDENOM_TAG=0xC0,
    EXPAND_TAG=0xC1, FACTOR_TAG=0xC2, CFACTOR_TAG=0xC3,
    INTEGRATE_TAG=0xC4, DIFFERENTIATE_TAG=0xC5,
    AVGRC_TAG=0xC6, NDERIV_TAG=0xC7, TAYLOR_TAG=0xC8,
    LIMIT_TAG=0xC9,
    PROPFRAC_TAG=0xCA,
    WHEN_TAG=0xCB,
    ROUND_TAG=0xCC,
    DMS_TAG=0xCD,
    LEFT_TAG=0xCE, RIGHT_TAG=0xCF, MID_TAG=0xD0,
    SHIFT_TAG=0xD1,
    SEQ_TAG=0xD2,
    LIST2MAT_TAG=0xD3, SUBMAT_TAG=0xD4,
    SUBSCRIPT_TAG=0xD5,
    RAND_TAG=0xD6,
    MIN_TAG=0xD7, MAX_TAG=0xD8,
    LIST_TAG=0xD9,
    USERFUNC_TAG=0xDA,
    V_AUGMENT_TAG=0xDB, // ?
    // Nothing ?
    // Nothing ?
    // ?
    // ?
    // Nothing ?
    // Nothing ?
    // Nothing ?
    // Nothing ?
    // Nothing ?
    END_TAG=0xE5,
    // Nothing ?
    // Nothing ?
    // Nothing ?
    // Nothing ?
    PN1_TAG=0xEA, PN2_TAG=0xEB,
    // ?
    EIGVC_TAG=0xED, EIGVL_TAG=0xEE,
    // ?
    LOCALVAR_TAG=0xF0, // probably
    DESOLVE_TAG=0xF1,
    FDASH_TAG=0xF2,
    ROTATE_TAG=0xF3,
    ISPRIME_TAG=0xF4,
    ////
    ORD_TAG=0xF8, CHAR_TAG=0xF9,
    STRING_TAG=0xFA, EXPR_TAG=0xFB,
    FORMAT_TAG=0xFC,
    INSTRING_TAG=0xFD,
    APPEND_TAG=0xFE,
    TMPCNV_TAG=0xFF, DELTA_TMPCNV_TAG=0x100,
    POLCPLX_TAG=0x101,
    // ?
    GRAD_TAG=0x103, // probably
    UNITCONV_TAG=0x104,
    ////
    TO_LOGBASE_TAG=0x112,
    ////
    ONEVAR_TAG=0x134,
    ////
    TWOVAR_TAG=0x139,
    ////
    CUBICREG_TAG=0x142, EXPREG_TAG=0x143, LINREGAX_TAG=0x144, LINREGBX_TAG=0x145, LNREG_TAG=0x146, LOGISTIC_TAG=0x147, LOGISTIC83_TAG=0x148, MEDMED_TAG=0x149, MULTREG_TAG=0x14A, POWERREG_TAG=0x14B, QUADREG_TAG=0x14C, QUARTREG_TAG=0x14D, SINREG_TAG=0x14E,
    // ?
    DELVAR_TAG=0x150,
    CYCLE_TAG=0x151,
    EXIT_TAG=0x152,
    FOR_TAG=0x153,
    ENDFOR_TAG=0x154,
    // ?
    ENDLOOP_TAG=0x156,
    WHILE_TAG=0x157,
    ENDWHILE_TAG=0x158,
    IF_TAG=0x159,
    // ENDIF_TAG ?
    IFTHEN_TAG=0x15B,
    ELSEIF_TAG=0x15C,
    ELSE_TAG=0x15D,
    // ?
    GOTO_TAG=0x15F,
    // ?
    // ?
    // ?
    RANDSEED_TAG=0x163,
    FILL_TAG=0x164,
    RETURN_TAG=0x165,
    LOCAL_TAG=0x166,
    SORTA_TAG=0x167, SORTD_TAG=0x168,
    LU_TAG=0x169, QR_TAG=0x16A,
    DEFINE_TAG=0x16B,
    ////
    COPYVAR_TAG=0x179,
    ////
    TRY_TAG=0x17E,
    // ?
    ENDTRY_TAG=0x180,
    CLRERR_TAG=0x181, PASSERR_TAG=0x182, ERRCODE_TAG=0x183,
    DISP_TAG=0x184,
    GETMODE_TAG=0x185, SETMODE_TAG=0x186,
    ////
    PART_TAG=0x1A1,
    GETTYPE_TAG=0x1A2, GETVARINFO_TAG=0x1A3,
    ////
    TANLINE_TAG=0x1AB, NORMLINE_TAG=0x1AC,
    ////
    LOCK_TAG=0x1B4, UNLOCK_TAG=0x1B5, ISLOCKED_TAG=0x1B6,
    ////
};

typedef struct {
    uint8_t tag_order;
    uint8_t tag_version;
    uint8_t min_max_args; /* 4 MSB for minimal number of arguments, 4 LSB for maximal number of arguments */
    uint8_t proc_args;
    void (*tag_proc) (); /* Routine that handles the tag */
    uint32_t tag_str; /* String ID */
    uint32_t unknown;
} tag_info;

// 441 elements in OS 1.7.2741 CAS
#define primary_tag_list ((const tag_info *)0x1069185C)

#define push_pi ((void (*)(void))(primary_tag_list[PI_TAG].tag_proc))
#define push_exp_1 ((void (*)(void))(primary_tag_list[EXP_TAG].tag_proc))
#define push_simplify_string ((void (*)(CESI))(primary_tag_list[STRING_TAG].tag_proc))


int main(void) {
    ESI old_top;
    ESI s0,s1,s2,s3,sa,sb,ptr;
    ESI ptr1, ptr2, ptr3, ptr4, ptr5, ptr6;
    unsigned int argcount;

    old_top = top_estack;
    printf("%p\n", old_top);
    ptr1 = old_top;

    push_quantum(END_TAG);

    // 100
    push_quantum_as_nonnegative_int(100);
    ptr1 = top_estack;
    // 1
    push_quantum_as_nonnegative_int(1);
    ptr2 = top_estack;
    // x
    push_quantum(0x08);
    ptr3 = top_estack;

    // x
    push_quantum(0x08);
    ptr4 = top_estack;
    // x
    push_quantum(0x08);
    // x^2
    replace_top2_with_prod(ptr4);
    ptr5 = top_estack;
    // 1 / x^2
    replace_top_with_reciprocal(top_estack);
    ptr6 = top_estack;

    // seq(1/x^2, x, 1, 100, 1, 1)
    push_sequence(ptr6, ptr3, ptr2, ptr1, Integer1Index);
    ptr1 = top_estack;

    ptr = top_estack;
    if (*ptr != LIST_TAG) {
        //ER_throw(ER_ARGUMENT_MUST_BE_LIST);
    }
    ptr--;
    if (*ptr == LIST_TAG) {
        //ER_throw(ER_ARGUMENT_MUST_BE_LIST);
    }

    if (remaining_element_count(ptr+1) > 1) {
        //ER_throw(ER_TOO_MANY_ARGS);
    }
    old_top = next_expression_index(ptr+1);
    argcount = remaining_element_count(ptr);

    if (argcount < 5) {
        //ER_throw(ER_DIMENSION);
    }

    // Skip LIST_TAG.
    top_estack--;

    argcount -= 4;
    while (argcount) {
        ptr = top_estack;
        replace_top2_with_sum(next_expression_index(ptr));
        argcount--;
    }
    
    // Now, there are only four elements left.
    // We must compute the partial sums before applying Aitken's delta^2
    // algorithm.
    s3 = top_estack;
    s2 = next_expression_index(s3);
    s1 = next_expression_index(s2);
    s0 = next_expression_index(s1);
    
    
    push_sum(s3,s2);
    s2=top_estack;
    push_sum(s2,s1);
    s1=top_estack;
    push_sum(s1,s0);
    s0=top_estack;
    
    // Before applying Aitken's delta^2 algorithm, check whether the three
    // last partial sums are equal (which means the series has 
    // converged to its sum).
    // Use the efficient way, size-wise, speed is not a real concern anyway.
    if (!compare_expressions(s1,s2)) {
        if (!compare_expressions(s1,s0)) {
            sb = s0;
            sa = s0;
            goto __push_result_and_return__;
        }
    }

    
    // First part of Aitken's delta^2 algorithm.
    // sb = s0-(s0-s1)^2/(s0-2*s1+s2).
    push_difference(s0,s1);
    times_top(top_estack);
    push_sum(s0,s2);
    subtract_from_top(s1);
    subtract_from_top(s1);
   
    replace_top2_with_ratio(next_expression_index(top_estack));
    push_difference(s0,top_estack);
    sb = top_estack;

    
    // Second part of Aitken's delta^2 algorithm.
    // sa = s1-(s1-s2)^2/(s1-2*s2+s3)
    push_difference(s1,s2);
    times_top(top_estack);
    
    push_sum(s1,s3);
    subtract_from_top(s2);
    subtract_from_top(s2);

    replace_top2_with_ratio(next_expression_index(top_estack));
    push_difference(s1,top_estack);
    sa = top_estack;
    

__push_result_and_return__:
    // Push result onto the EStack, cleanup.
    push_quantum(END_TAG);
    push_difference(sb,sa);
    push_expression(sb);
    push_quantum(LIST_TAG);
    delete_between(old_top,next_expression_index(top_estack));

    ptr1 = top_estack + 1;

    // string(result)
    push_simplify_string(top_estack);

    // Print the bytes of the string on the RS232 port.
    for (; ptr1 <= top_estack; ptr1++) {
        printf("%c", *(uint8_t *)ptr1);
    }
    printf("\n");

    return 0;
}
